#include "flpydsk.h"
// detailed infos about FDC and FAT12:
// http://www.isdaman.com/alsos/hardware/fdc/floppy.htm
// http://lowlevel.brainsware.org/wiki/index.php/FDC#Data_Address_Mark
// http://www.brokenthorn.com/Resources/OSDev20.html
// http://www.eit.lth.se/fileadmin/eit/courses/eit015/FAT12Description.pdf

// IO ports
enum FLPYDSK_IO
{
	FLPYDSK_DOR		=	0x3f2,
	FLPYDSK_MSR		=	0x3f4,
	FLPYDSK_FIFO	=	0x3f5,
	FLPYDSK_CTRL	=	0x3f7
};

// Bits 0-4 of command byte.
enum FLPYDSK_CMD
{
	FDC_CMD_READ_TRACK	 =	  2,
	FDC_CMD_SPECIFY		 =	  3,
	FDC_CMD_CHECK_STAT	 =	  4,
	FDC_CMD_WRITE_SECT	 =	  5,
	FDC_CMD_READ_SECT	 =	  6,
	FDC_CMD_CALIBRATE	 =	  7,
	FDC_CMD_CHECK_INT	 =	  8,
	FDC_CMD_FORMAT_TRACK =	0xd,
	FDC_CMD_SEEK		 =	0xf
};

// Additional command masks. Can be masked with above commands
enum FLPYDSK_CMD_EXT
{
	FDC_CMD_EXT_SKIP		=	0x20,	//00100000
	FDC_CMD_EXT_DENSITY		=	0x40,	//01000000
	FDC_CMD_EXT_MULTITRACK	=	0x80	//10000000
};

// Digital Output Register
enum FLPYDSK_DOR_MASK
{
	FLPYDSK_DOR_MASK_DRIVE0			=	0,	//00000000
	FLPYDSK_DOR_MASK_DRIVE1			=	1,	//00000001
	FLPYDSK_DOR_MASK_DRIVE2			=	2,	//00000010
	FLPYDSK_DOR_MASK_DRIVE3			=	3,	//00000011
	FLPYDSK_DOR_MASK_RESET			=	4,	//00000100
	FLPYDSK_DOR_MASK_DMA			=	8,	//00001000
	FLPYDSK_DOR_MASK_DRIVE0_MOTOR	=	16,	//00010000
	FLPYDSK_DOR_MASK_DRIVE1_MOTOR	=	32,	//00100000
	FLPYDSK_DOR_MASK_DRIVE2_MOTOR	=	64,	//01000000
	FLPYDSK_DOR_MASK_DRIVE3_MOTOR	=	128	//10000000
};

// Main Status Register
enum FLPYDSK_MSR_MASK
{
	FLPYDSK_MSR_MASK_DRIVE1_POS_MODE	=	1,	//00000001
	FLPYDSK_MSR_MASK_DRIVE2_POS_MODE	=	2,	//00000010
	FLPYDSK_MSR_MASK_DRIVE3_POS_MODE	=	4,	//00000100
	FLPYDSK_MSR_MASK_DRIVE4_POS_MODE	=	8,	//00001000
	FLPYDSK_MSR_MASK_BUSY				=	16,	//00010000
	FLPYDSK_MSR_MASK_DMA				=	32,	//00100000
	FLPYDSK_MSR_MASK_DATAIO				=	64, //01000000
	FLPYDSK_MSR_MASK_DATAREG			=	128	//10000000
};

// Controller Status Port 0
enum FLPYDSK_ST0_MASK
{
	FLPYDSK_ST0_MASK_DRIVE0		=	0,		//00000000
	FLPYDSK_ST0_MASK_DRIVE1		=	1,		//00000001
	FLPYDSK_ST0_MASK_DRIVE2		=	2,		//00000010
	FLPYDSK_ST0_MASK_DRIVE3		=	3,		//00000011
	FLPYDSK_ST0_MASK_HEADACTIVE	=	4,		//00000100
	FLPYDSK_ST0_MASK_NOTREADY	=	8,		//00001000
	FLPYDSK_ST0_MASK_UNITCHECK	=	16,		//00010000
	FLPYDSK_ST0_MASK_SEEKEND	=	32,		//00100000
	FLPYDSK_ST0_MASK_INTCODE	=	64		//11000000
};

// LPYDSK_ST0_MASK_INTCODE types
enum FLPYDSK_ST0_INTCODE_TYP
{
	FLPYDSK_ST0_TYP_NORMAL		 =	0,
	FLPYDSK_ST0_TYP_ABNORMAL_ERR =	1,
	FLPYDSK_ST0_TYP_INVALID_ERR	 =	2,
	FLPYDSK_ST0_TYP_NOTREADY	 =	3
};

// GAP 3 sizes
enum FLPYDSK_GAP3_LENGTH
{
	FLPYDSK_GAP3_LENGTH_STD  = 42,
	FLPYDSK_GAP3_LENGTH_5_14 = 32,
	FLPYDSK_GAP3_LENGTH_3_5  = 27
};

// Formula: 2^sector_number * 128
enum FLPYDSK_SECTOR_DTL
{

	FLPYDSK_SECTOR_DTL_128	=	0,
	FLPYDSK_SECTOR_DTL_256	=	1,
	FLPYDSK_SECTOR_DTL_512	=	2,
	FLPYDSK_SECTOR_DTL_1024	=	4
};

/**********************************
* Constants
**********************************/

const int FLOPPY_IRQ                         =  6;     // floppy irq
const int FLPY_SECTORS_PER_TRACK             = 18;     // sectors per track
static unsigned char	_CurrentDrive        =  0;     // current working drive. Defaults to 0.
static volatile unsigned char _FloppyDiskIRQ =  0;     // set when IRQ fires


/**
**	DMA Routines.
**	The DMA (Direct Memory Access) controller allows the FDC to send data to the DMA, which can put the data in memory.
**  While the FDC can be programmed to not use DMA, it is not very well supported on emulators or virtual machines.
**  Because of this, the DMA is used for data transfers.
**/

// initialize DMA to use physical address 84k-128k
void flpydsk_initialize_dma()
{
	outportb(0x0a, 0x06);	//mask dma channel 2
	outportb(0xd8, 0xFF);	//reset master flip-flop
	outportb(0x04, 0x00);   //address=0x1000
	outportb(0x04, 0x10);
	outportb(0xd8, 0xFF);   //reset master flip-flop
	outportb(0x05, 0xFF);   //count to 0x23FF (number of bytes in a 3.5" floppy disk track: 18*512)
	outportb(0x05, 0x23);

	outportb(0x81, 0x00);   //external page register = 0
	outportb(0x0a, 0x02);   //unmask dma channel 2
}

// prepare the DMA for read transfer
void flpydsk_dma_read()
{
	outportb(0x0a, 0x06); //mask dma channel 2
	outportb(0x0b, 0x56); //single transfer, address increment, autoinit, read, channel 2
	outportb(0x0a, 0x02); //unmask dma channel 2
}

// prepare the DMA for write transfer
void flpydsk_dma_write()
{
	outportb(0x0a, 0x06); //mask dma channel 2
	outportb(0x0b, 0x5A); //single transfer, address increment, autoinit, write, channel 2
	outportb(0x0a, 0x02); //unmask dma channel 2
}


/**
*	Basic Controller In/Out Routines
*/

// return fdc status
unsigned char flpydsk_read_status ()
{
	return inportb(FLPYDSK_MSR); // just return main status register
}

// write to the fdc dor
void flpydsk_write_dor( unsigned char val )
{
	outportb(FLPYDSK_DOR, val); // write the digital output register
}

// send command byte to fdc
void flpydsk_send_command (unsigned char cmd)
{
    // wait until data register is ready. We send commands to the data register
	int i;
	for(i=0; i<500; ++i)
		if( flpydsk_read_status() & FLPYDSK_MSR_MASK_DATAREG )
			return outportb(FLPYDSK_FIFO, cmd);
}

// get data from fdc
unsigned char flpydsk_read_data()
{
	// same as above function but returns data register for reading
	int i;
	for(i=0; i<500; ++i)
		if( flpydsk_read_status() & FLPYDSK_MSR_MASK_DATAREG )
			return inportb(FLPYDSK_FIFO);
	return 0;
}

// write to the configuation control register
void flpydsk_write_ccr(unsigned char val)
{
	outportb(FLPYDSK_CTRL, val); // write the configuation control
}

/**
*	Interrupt Handling Routines
*/

// wait for irq
inline void flpydsk_wait_irq()
{
    while ( _FloppyDiskIRQ == 0) // wait for irq to fire
		;
	_FloppyDiskIRQ = 0;
}

//	floppy disk irq handler
void i86_flpy_irq(struct regs* r)
{
	_FloppyDiskIRQ = 1; // irq fired. Set flag!
	printformat("DEBUG flpydsk.c: after _FloppyDiskIRQ = 1; // irq fired. Set flag!\n");
}

/**
*	Controller Command Routines
*/

// check interrupt status command
void flpydsk_check_int(ULONG* st0, ULONG* cyl)
{
	flpydsk_send_command(FDC_CMD_CHECK_INT);
	*st0 = flpydsk_read_data();
	*cyl = flpydsk_read_data();
}

// turns the current floppy drives motor on/off
void flpydsk_control_motor(int b)
{
	// sanity check: invalid drive
	if(_CurrentDrive > 3) return;
	ULONG motor = 0;

	switch(_CurrentDrive) // select the correct mask based on current drive
	{
		case 0:
			motor = FLPYDSK_DOR_MASK_DRIVE0_MOTOR; break;
		case 1:
			motor = FLPYDSK_DOR_MASK_DRIVE1_MOTOR; break;
		case 2:
			motor = FLPYDSK_DOR_MASK_DRIVE2_MOTOR; break;
		case 3:
			motor = FLPYDSK_DOR_MASK_DRIVE3_MOTOR; break;
	}
	// turn on or off the motor of that drive
	if(b)
		flpydsk_write_dor(_CurrentDrive | motor | FLPYDSK_DOR_MASK_RESET | FLPYDSK_DOR_MASK_DMA);
	else
		flpydsk_write_dor(FLPYDSK_DOR_MASK_RESET);
	sleepMilliSeconds(20); // in all cases: wait a little bit for the motor to spin up/turn off
}

// configure drive
void flpydsk_drive_data(ULONG stepr, ULONG loadt, ULONG unloadt, int dma )
{
	ULONG data = 0;
	flpydsk_send_command (FDC_CMD_SPECIFY); // send command
	data = ((stepr & 0xf) << 4) | (unloadt & 0xf);
	flpydsk_send_command(data);
	data = (loadt) << 1 | (dma==FALSE) ? 0 : 1;
	flpydsk_send_command(data);
}

// calibrate the drive
int flpydsk_calibrate(ULONG drive)
{
	ULONG st0, cyl;
	if (drive >= 4)	return -2;
	flpydsk_control_motor(TRUE); // turn on the motor
	int i;
	for(i=0; i<10; ++i)
	{
		// send command
		flpydsk_send_command( FDC_CMD_CALIBRATE );
		flpydsk_send_command( drive );
		flpydsk_wait_irq();
		flpydsk_check_int(&st0, &cyl);

		if(!cyl) // did we find cylinder 0? if so, we are done
		{
			flpydsk_control_motor(FALSE);
			return 0;
		}
	}
	flpydsk_control_motor(FALSE);
	return -1;
}

// disable controller
void flpydsk_disable_controller()
{
	flpydsk_write_dor(0);
}

// enable controller
void flpydsk_enable_controller()
{
	flpydsk_write_dor(FLPYDSK_DOR_MASK_RESET | FLPYDSK_DOR_MASK_DMA);
}

// reset controller
void flpydsk_reset()
{
	ULONG st0, cyl;
	flpydsk_disable_controller();
	flpydsk_enable_controller();
	flpydsk_wait_irq();

	// send CHECK_INT/SENSE INTERRUPT command to all drives
	int i;
	for(i=0; i<4; ++i)
		flpydsk_check_int(&st0,&cyl);
	flpydsk_write_ccr(0);              // transfer speed 500kb/s
	flpydsk_drive_data(3,16,0xF,TRUE); // pass mechanical drive info: steprate=3ms, load time=16ms, unload time=240ms (0xF bei 500K)
	flpydsk_calibrate(_CurrentDrive);  // calibrate the disk
}

// seek to given track/cylinder
int flpydsk_seek( ULONG cyl, ULONG head )
{
	ULONG st0, cyl0;
	if (_CurrentDrive >= 4)	return -1;
	int i;
	for(i=0; i<10; ++i)
	{
		// send the command
		flpydsk_send_command (FDC_CMD_SEEK);
		flpydsk_send_command ( (head) << 2 | _CurrentDrive);
		flpydsk_send_command (cyl);

		flpydsk_wait_irq();
		flpydsk_check_int(&st0,&cyl0);

		if ( cyl0 == cyl) // found the cylinder?
			return 0;
	}
	return -1;
}

/*
http://en.wikipedia.org/wiki/CHS_conversion#From_LBA_to_CHS
CYL = LBA / (HPC * SPT)
TEMP = LBA % (HPC * SPT)
HEAD = TEMP / SPT
SECT = TEMP % SPT + 1
*/

// convert LBA to CHS
void flpydsk_lba_to_chs(int lba, int* head, int* track, int* sector)
{
   *track  =   lba / ( FLPY_SECTORS_PER_TRACK * 2 );
   *head   = ( lba % ( FLPY_SECTORS_PER_TRACK * 2 ) ) / ( FLPY_SECTORS_PER_TRACK );
   *sector = ( lba % ( FLPY_SECTORS_PER_TRACK * 2 ) ) % FLPY_SECTORS_PER_TRACK + 1;
}

/*
void flpydsk_lba_to_chs(int lba, int* head, int* track, int* sector) // also correct, more efficient
{
   *track  =   lba / ( FLPY_SECTORS_PER_TRACK * 2 );
   *head   = ( lba % ( FLPY_SECTORS_PER_TRACK * 2 ) ) / ( FLPY_SECTORS_PER_TRACK );
   *sector =   lba %   FLPY_SECTORS_PER_TRACK + 1;
}
*/

// install floppy driver
void flpydsk_install(int irq)
{
	irq_install_handler(irq, i86_flpy_irq);
	flpydsk_initialize_dma();
	flpydsk_reset();
	flpydsk_drive_data(13, 1, 0xF, TRUE);
}

// set current working drive
void flpydsk_set_working_drive(unsigned char drive){ if (drive < 4) _CurrentDrive = drive; }

// get current working drive
unsigned char flpydsk_get_working_drive(){ return _CurrentDrive; }

// read or write a sector
// http://www.isdaman.com/alsos/hardware/fdc/floppy_files/wrsec.gif
void flpydsk_transfer_sector(unsigned char head, unsigned char track, unsigned char sector, unsigned char operation) //read: operation = 0; write: operation = 1
{
	ULONG st0, cyl;

	if(operation == 0) // read a sector
	{
	    flpydsk_dma_read();
	    flpydsk_send_command( FDC_CMD_READ_SECT | FDC_CMD_EXT_MULTITRACK | FDC_CMD_EXT_SKIP | FDC_CMD_EXT_DENSITY);
	}
	if(operation == 1) // write a sector
	{
	    flpydsk_dma_write();
	    flpydsk_send_command( FDC_CMD_WRITE_SECT | FDC_CMD_EXT_MULTITRACK | FDC_CMD_EXT_DENSITY );
	}

	flpydsk_send_command( head << 2 | _CurrentDrive );
	flpydsk_send_command( track);
	flpydsk_send_command( head);
	flpydsk_send_command( sector);
	flpydsk_send_command( FLPYDSK_SECTOR_DTL_512 );
	flpydsk_send_command( FLPY_SECTORS_PER_TRACK );
	flpydsk_send_command( FLPYDSK_GAP3_LENGTH_3_5 );
	flpydsk_send_command( 0xFF );

	flpydsk_wait_irq();

	printformat("Return Values:\n");
	int j;
	for(j=0; j<7; ++j)
	{
		int val = flpydsk_read_data(); // read status info: ST0 ST1 ST2 C H S Size(2: 512 Byte)
		printformat("%d: %d  ",j,val);
	}
	printformat("\n\n");
	flpydsk_check_int(&st0,&cyl); // let FDC know we handled interrupt
	sleepSeconds(5);
}

// read a sector
unsigned char* flpydsk_read_sector(int sectorLBA)
{
	if (_CurrentDrive >= 4) return 0;
	// convert LBA sector to CHS
	int head=0, track=0, sector=1;
	flpydsk_lba_to_chs(sectorLBA, &head, &track, &sector);
	// turn motor on and seek to track
	flpydsk_control_motor(TRUE);
	if(flpydsk_seek (track, head)) return 0;
	// read sector and turn motor off
	flpydsk_transfer_sector(head, track, sector, 0);
	flpydsk_control_motor(FALSE);

	return (unsigned char*)DMA_BUFFER;
}

// write a sector
int flpydsk_write_sector(int sectorLBA)
{
	if (_CurrentDrive >= 4) return -1;
	// convert LBA sector to CHS
	int head=0, track=0, sector=1;
	flpydsk_lba_to_chs(sectorLBA, &head, &track, &sector);
	// turn motor on and seek to track
	flpydsk_control_motor(TRUE);
	if(flpydsk_seek (track, head)) return -2;
	// write sector and turn motor off
	flpydsk_transfer_sector(head, track, sector, 1);
	flpydsk_control_motor(FALSE);
	return 0;
}

void flpydsk_read_directory()
{
	printformat("DEBUG flpydsk.c: flpydsk_read_directory\n");
	flpydsk_read_sector(19); // start at 0x2600: root directory (14 sectors)
	printformat("<Floppy Disc Root Dir>\n");

	ULONG i;
	for(i=0;i<224;++i)       // 224 Entries * 32 Byte
	{
        if(
			(( *((unsigned char*)(DMA_BUFFER + i*32)) )      != 0x00 ) && /* free from here on           */
			(( *((unsigned char*)(DMA_BUFFER + i*32)) )      != 0xE5 ) && /* 0xE5 deleted = free         */
			(( *((unsigned char*)(DMA_BUFFER + i*32 + 11)) ) != 0x0F )    /* 0x0F part of long file name */
		  )
		  {
			int start = DMA_BUFFER + i*32; // name
			int count = 8;
			char* end = (char*)(start+count);
			for(; count != 0; --count)
			{
			    if( *(end-count) != 0x20 ) /* empty space in file name */
				    printformat("%c",*(end-count));
			}

			if(i!=0) printformat("."); // usual separator between file name and file extension

			start = DMA_BUFFER + i*32 + 8; // extension
			count = 3;
			end = (char*)(start+count);
			for(; count != 0; --count)
				printformat("%c",*(end-count));

			// filesize
			printformat("\t%d byte", *((ULONG*)(DMA_BUFFER + i*32 + 28)));

            // attributes
			printformat("\t");
			if(*((ULONG*)(DMA_BUFFER + i*32 + 28))<100) printformat("\t");
			if((( *((unsigned char*)(DMA_BUFFER + i*32 + 11)) ) & 0x08 ) == 0x08 ) printformat(" (lab)");
			if((( *((unsigned char*)(DMA_BUFFER + i*32 + 11)) ) & 0x10 ) == 0x10 ) printformat(" (dir)");
			if((( *((unsigned char*)(DMA_BUFFER + i*32 + 11)) ) & 0x01 ) == 0x01 ) printformat(" (r/o)");
			if((( *((unsigned char*)(DMA_BUFFER + i*32 + 11)) ) & 0x02 ) == 0x02 ) printformat(" (hid)");
			if((( *((unsigned char*)(DMA_BUFFER + i*32 + 11)) ) & 0x04 ) == 0x04 ) printformat(" (sys)");
			if((( *((unsigned char*)(DMA_BUFFER + i*32 + 11)) ) & 0x20 ) == 0x20 ) printformat(" (arc)");

			// 1st cluster: physical sector number  =  33  +  FAT entry number  -  2  =  FAT entry number  +  31
            printformat("  1st phys. sec: %d", *((USHORT*)(DMA_BUFFER + i*32 + 26))+31);

            printformat("\n"); // next root dirextory entry
		  }//if
	}//for
    printformat("\n");
}
